from json import dumps

import django
from django.test import TestCase

from perestroika.exceptions import BadRequest
from perestroika.utils import dict_to_multi_dict

django.setup()

from django.contrib.auth.models import User


class DjangoTest(TestCase):
    def make_post(self, url, data):
        return self.client.post(url, dumps(data), content_type='application/json')

    def make_empty_post(self, url):
        return self.client.post(url, content_type='application/json')

    def make_get(self, url, data):
        return self.client.get(url, dict_to_multi_dict(data), content_type='application/json')

    def test_allowed_methods(self):
        _response = self.make_post("/test/empty/", {})
        assert _response.status_code == 405

    def test_empty_get(self):
        User(username="first").save()

        assert User.objects.count() == 1

        _response = self.make_get("/test/full/", {})
        assert _response.status_code == 200
        assert _response.json() == {
            'item': {'username': "first"},
            'limit': 20,
            'skip': 0,
            'status_code': 200,
            'total_count': 1
        }

        User(username="second").save()

        assert User.objects.count() == 2

        _response = self.make_get("/test/full/", {})
        assert _response.status_code == 200
        assert _response.json() == {
            'items': [{'username': "first"}, {"username": "second"}],
            'limit': 20,
            'skip': 0,
            'status_code': 200,
            'total_count': 2
        }

    def test_json_validation_no_items(self):
        with self.assertRaises(BadRequest):
            _response = self.make_empty_post("/test/full/")



from typing import Dict, Union, Optional

import attr

from perestroika.methods import Method


@attr.s
class DjangoResource:
    from django.views.decorators.csrf import csrf_exempt

    methods: Optional[Dict[str, Method]] = None
    cache_control: Dict[str, Union[bool, int]] = attr.Factory(dict)

    @csrf_exempt
    def handler(self, request):
        from django.http import HttpResponseNotAllowed
        from django.utils.cache import patch_cache_control

        if self.methods:
            method = self.methods.get(request.method.lower())

            if method:
                response = method.handle(request)

                if self.cache_control:
                    patch_cache_control(response, **self.cache_control)

                return response

        permitted_methods = self.methods.keys() if self.methods else []
        return HttpResponseNotAllowed(permitted_methods=permitted_methods)

    def schema(self, request):
        from django.http import JsonResponse
        _schema = {
            k: v.schema() for k, v in self.methods.items()
        }

        return JsonResponse(_schema)



from django.urls import path

from tests.test_django.resources import EmptyResource, FullResource

urlpatterns = [
    path('empty/', EmptyResource().handler),
    path('full/', FullResource().handler),
]



from perestroika.exceptions import BadRequest


class DbLayer:
    pass


class DjangoDbLayer(DbLayer):
    @staticmethod
    def get(bundle, method):
        _filter = bundle.get("filter")
        _exclude = bundle.get("exclude")
        _project = bundle.get("project")

        if _filter:
            bundle["queryset"] = bundle["queryset"].filter(_filter)

        if _exclude:
            bundle["queryset"] = bundle["queryset"].exclude(_exclude)

        if _project:
            bundle["queryset"] = bundle["queryset"].only(*_project)

        bundle["items"] = bundle["queryset"].values()
        bundle["total_count"] = bundle["queryset"].count()

    @staticmethod
    def post(bundle, method):
        _objects = bundle.get("items")

        if not _objects:
            raise BadRequest(message="Empty data for resource")

        _objects = [bundle["queryset"].model(**data) for data in _objects]

        bundle["queryset"].model.objects.bulk_create(
            _objects
        )



import logging
from collections import Callable
from typing import List, Any, Optional

import attr
from validate_it import Schema

from perestroika.db_layers import DbLayer, DjangoDbLayer
from perestroika.deserializers import Deserializer, DjangoDeserializer
from perestroika.exceptions import RestException, BadRequest, InternalServerError
from perestroika.serializers import Serializer, DjangoSerializer

logger = logging.getLogger(__name__)


class DenyAll(Schema):
    def __init__(self, **kwargs) -> None:
        raise TypeError("Deny all types")


@attr.s(auto_attribs=True)
class Method:
    mode: str = 'django'

    query: Any = None
    queryset: Any = None
    db_layer: Optional[DbLayer] = None

    serializer: Optional[Serializer] = None
    deserializer: Optional[Deserializer] = None

    skip_query_db: bool = False

    input_validator: type = DenyAll
    output_validator: type = DenyAll

    pre_query_hooks: List[Callable] = attr.Factory(list)
    post_query_hooks: List[Callable] = attr.Factory(list)

    request_hooks: List[Callable] = attr.Factory(list)
    response_hooks: List[Callable] = attr.Factory(list)

    def __attrs_post_init__(self):
        need_fields = []

        if self.mode == 'django':
            need_fields = ['queryset']

            # allow custom db layers
            if not isinstance(self.db_layer, DjangoDbLayer):
                self.db_layer = DjangoDbLayer()

            # allow custom serializers
            if not isinstance(self.serializer, DjangoSerializer):
                self.serializer = DjangoSerializer()

            # allow custom deserializers
            if not isinstance(self.deserializer, DjangoDeserializer):
                self.deserializer = DjangoDeserializer()

        for field in need_fields:
            if getattr(self, field) is None and not self.skip_query_db:
                raise ValueError(f"Empty `{field}` is allowed only for resources with `skip_query_db` == True")

    def __set_name__(self, owner, name):
        if not owner.methods:
            owner.methods = {}

        owner.methods[self.__class__.__name__.lower()] = self

    def schema(self):
        return {
            "output_schema": self.output_validator.representation()
        }

    def get_client_data(self, request):
        return self.deserializer.deserialize(request, self)

    def query_db(self, bundle):
        raise NotImplementedError()

    @staticmethod
    def validate(validator: Schema, bundle, validation_exception_class=RestException):
        _errors = []
        _objects = []

        for _object in bundle["items"]:
            try:
                _object = validator.from_dict(_object).to_dict()
                _objects.append(_object)
            except TypeError as e:
                _errors.append(e)

        if _errors:
            raise validation_exception_class(message={"errors": _errors, "items": bundle["items"]})

        bundle["items"] = _objects

    def validate_input(self, bundle):
        self.validate(self.input_validator, bundle, validation_exception_class=BadRequest)

    def validate_output(self, bundle):
        self.output_validator.Meta.strip_unknown = True
        self.validate(self.output_validator, bundle, validation_exception_class=InternalServerError)

    @staticmethod
    def apply_hooks(hooks, request, bundle):
        for hook in hooks:
            hook(request, bundle)

    def apply_pre_query_hooks(self, request, bundle):
        self.apply_hooks(self.pre_query_hooks, request, bundle)

    def apply_post_query_hooks(self, request, bundle):
        self.apply_hooks(self.post_query_hooks, request, bundle)

    def set_default_success_code(self, bundle):
        raise NotImplementedError()

    def apply_request_hooks(self, request, bundle):
        self.apply_hooks(self.request_hooks, request, bundle)

    def apply_response_hooks(self, request, bundle):
        self.apply_hooks(self.response_hooks, request, bundle)

    def handle(self, request):
        bundle = self.get_client_data(request)

        self.set_default_success_code(bundle)
        self.apply_request_hooks(request, bundle)
        self.validate_input(bundle)
        self.apply_pre_query_hooks(request, bundle)

        if not self.skip_query_db:
            self.query_db(bundle)

        self.apply_post_query_hooks(request, bundle)
        self.validate_output(bundle)
        self.apply_response_hooks(request, bundle)

        return self.serializer.serialize(request, bundle)


@attr.s(auto_attribs=True)
class CanFilterAndExclude(Method):
    filter_validator: type = DenyAll
    exclude_validator: type = DenyAll

    def set_default_success_code(self, bundle):
        raise NotImplementedError()

    def query_db(self, bundle):
        raise NotImplementedError()

    def schema(self):
        _schema = super().schema()
        _schema["filter_schema"] = self.filter_validator.representation()
        _schema["exclude_schema"] = self.exclude_validator.representation()
        return _schema


@attr.s(auto_attribs=True)
class NoBodyNoObjectsNoInput(CanFilterAndExclude):
    def set_default_success_code(self, bundle):
        raise NotImplementedError()

    def validate_input(self, bundle):
        """ Void validation because no input data"""
        pass

    def query_db(self, bundle):
        raise NotImplementedError()


@attr.s(auto_attribs=True)
class Get(NoBodyNoObjectsNoInput):
    def query_db(self, bundle):
        self.db_layer.get(bundle, self)

    def set_default_success_code(self, bundle):
        bundle["status_code"] = 200


@attr.s(auto_attribs=True)
class Post(Method):
    input_validator: type = DenyAll

    def query_db(self, bundle):
        self.db_layer.post(bundle, self)

    def schema(self):
        _schema = super().schema()
        _schema["input_schema"] = self.input_validator.representation()
        return _schema

    def set_default_success_code(self, bundle):
        bundle["status_code"] = 201


@attr.s(auto_attribs=True)
class Put(CanFilterAndExclude):
    def query_db(self, bundle):
        raise NotImplementedError()

    input_validator: type = DenyAll

    def set_default_success_code(self, bundle):
        raise NotImplementedError()

    def schema(self):
        _schema = super().schema()
        _schema["input_schema"] = self.input_validator.representation()
        return _schema


@attr.s(auto_attribs=True)
class Patch(CanFilterAndExclude):
    input_validator: type = DenyAll

    def query_db(self, bundle):
        raise NotImplementedError()

    def set_default_success_code(self, bundle):
        raise NotImplementedError()

    def schema(self):
        _schema = super().schema()
        _schema["input_schema"] = self.input_validator.representation()
        return _schema


@attr.s(auto_attribs=True)
class Delete(NoBodyNoObjectsNoInput):
    def query_db(self, bundle):
        raise NotImplementedError()

    def set_default_success_code(self, bundle):
        raise NotImplementedError()



from django.core.serializers.json import DjangoJSONEncoder
from django.http import JsonResponse


class Serializer:
    pass


class DjangoSerializer(Serializer):
    @staticmethod
    def get_encoder():
        return DjangoJSONEncoder

    def serialize(self, request, bundle):
        _data = {
            "limit": bundle.get("limit", 0),
            "skip": bundle.get("skip", 0),
            "total_count": bundle.get("total_count", 0),
        }

        if bundle.get("project"):
            _new_items = []

            for item in bundle["items"]:
                _new_item = {}

                for key in bundle["project"]:
                    _new_item[key] = item[key]

                _new_items.append(_new_item)

            bundle["items"] = _new_items

        _items = bundle.get("items", [])

        if len(_items) is 1:
            _data["item"] = _items[0]
        else:
            _data["items"] = _items

        for item in ['filter', 'order', 'project', 'error_code', 'error_message', 'status_code']:
            if bundle.get(item):
                _data[item] = bundle[item]

        return JsonResponse(_data, status=bundle['status_code'], encoder=self.get_encoder())



class RestException(Exception):
    def __init__(self, message=None, status_code=None) -> None:
        self.message = message
        self.status_code = status_code

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"


class BadRequest(RestException):
    def __init__(self, message=None) -> None:
        super().__init__(message, 400)


class InternalServerError(RestException):
    def __init__(self, message=None) -> None:
        super().__init__(message, 500)


class MethodNotAllowed(RestException):
    def __init__(self, message=None) -> None:
        super().__init__(message, 405)



import json

from perestroika.exceptions import BadRequest
from perestroika.utils import multi_dict_to_dict


class Deserializer:
    def deserialize(self, request, method):
        raise NotImplementedError()


class DjangoDeserializer(Deserializer):
    def deserialize(self, request, method):
        if request.method == 'GET':
            _data = multi_dict_to_dict(request.GET)
        else:
            _data = json.loads(request.body)

        _items = _data.get("items", [])
        _item = _data.get("item")

        if _item:
            _items = [_item]

        if not _items and request.method in ["POST", "PUT", "PATCH"]:
            raise BadRequest(message="Need data for processing")

        bundle = {
            "order": _data.get("order"),
            "filter": _data.get("filter"),
            "items": _items,
            "limit": 20,
            "skip": 0,
            "total_count": 0,
            "queryset": method.queryset,
            "project": _data.get("project", []),
        }

        return bundle
